#!/usr/bin/perl

$version = "#VERSION#";

$GCC = "gcc";
$BISON = "#BISON#";

sub Usage {
    print STDERR 'Usage:
      jbison [ --output=OUTPUT ] [ -v ] [ --version ] [-h] FILE.EXT

   Type "jbison -h" for help.
';
    exit 1;
}

sub Help {
   print STDERR 'Usage:
      jbison [ --output=OUTPUT ] [ -v ] [ --version ] [-h] [DIR/]FILE.EXT

   Produces a source file for a Java class implementing a LALR(1)
   parser described by the Bison input file FILE.EXT.  Here, EXT is
   an extension (containing no periods), typically "y" or "jy".  The
   resulting source file goes to OUTPUT (default FILE.java).  The -v
   switch produces the standard Bison .output file, and --version prints
   version information and exits.

   Jbison implements several additional directives, which you may include
   in the first declaration section of FILE.EXT: 

        %header{ 
           LINES
        %}
            Use LINES to introduce the class.  These should include any
            "package" or "import" clauses needed as well as the part of the
            class declaration preceding "{"  By default, uses
	        public class FILE 
        %apiaccess ACCESS
            Use ACCESS as the access level (private, default, 
            protected, public) of the function parse.
        %parserdecl SIGNATURE
            Use SIGNATURE as the header of function parse.  The
            default is "int parse ()"
        %valuetype TYPE
            Use TYPE as the type of semantic values (default is int).
';
   exit 0;
}

$output="";
$verbose="";
$input="";

for (@ARGV) {
    if (/^--output=(.*)/) {
        $output=$1;
    } elsif (/^-v$/ or /^--verbose$/) {
        $verbose = " -v";
    } elsif (/^--version$/ or /^-V$/) {
        print "$version\n"; exit 0;
    } elsif (/^-/ or $input) {
        Usage;
    } elsif (/^-h/) {
        Help;
    } else {
	m{(.*/)?([^/]*)\.[^.]*};
        $input = $_;
	$root = $2;
    }
}

if (not $input or not $root) {
    Usage;
}   

if (not $output) {
   $output="${root}.java";
}
 
open IN, "<$input" or die "Could not open file $input: $!\n";
$tmp1 = "${root}.y.tmp1";
unlink $tmp1;
open OUT, ">$tmp1" or die "Could not open temporary file $tmp1: $!\n";
$header = "";
$inHeader = 0;
$apiaccess = "public";
$valuetype = "int";
$parserdecl = "int parse ()";
while (<IN>) {
    if (/^%%/) {
	print OUT $_;
	while (<IN>) {
	    print OUT $_;
        }
        last;
    }  
    if ($inHeader) {
	if (/^\s*%}/) {
	    $inHeader = 0;
        } else {
            $header .= $_;
        }
        $_ = "\n"
    } elsif (/^\s*%header\s*{\s*$/) {
	$inHeader = 1;
	$_ = "\n";
    } elsif (/^\s*\%apiaccess\s+(.*?)\s*$/) {
	$apiaccess = $1;
	$_ = "\n";
    } elsif (/^\s*\%parserdecl\s+(.*?)\s*$/) {
	$parserdecl = $1;
	$_ = "\n";
    } elsif (/^\s*\%valuetype\s+(.*?)\s*$/) {
	$valuetype = $1;
	$_ = "\n";
    }
    print OUT $_;
}
close IN;
close OUT;

$tmp2 = "$output.tmp2";
$tmp3 = "$output.tmp3";
unlink $tmp2, $tmp3;
$code = system ("$BISON --output=$tmp2$verbose --skeleton=yacc.java $tmp1 2>$tmp3");
unlink $tmp1;
if (open MSGS, $tmp3) {
    while (<MSGS>) {
	s/^(.*)\.tmp1:(\d+)(\.(-|\d)+)?:/$1:$2:/;
        print STDERR $_;
    }
    close MSGS;
    unlink $tmp3;
}
if ($code != 0) {
   unlink $tmp2;
   exit 1;
}

open IN, "<$tmp2" or die "Could not open Bison output file: $!\n";
unlink $output;
open OUT, "|$GCC -E -C -P -undef -o $output -";

if (not $header) {
    $header = "public class $root\n";
}

while (<IN>) {
   if (/\%HEADER\%/) {
       print OUT $header;
   } else {
       s,^(// _line_ [0-9]+\s+")$tmp1",$1$input",;
       s,^(// _line_ [0-9]+\s+")$tmp2",$1$output",;
       s/\%APIACCESS\%/$apiaccess/g;
       s/\%PARSERDECL\%/$parserdecl/g;
       s/\%SEMANTIC\%/$valuetype/g;
       print OUT $_;
   }
}

close IN;
unlink $tmp2;
close OUT or die "Final pre-processing step failed.\n";

exit 0;
